/*
 * Copyright (c) 2003, 2004, 2005
 * The Regents of The University of Michigan
 * All Rights Reserved
 *
 * This code is part of the M5 simulator, developed by Nathan Binkert,
 * Erik Hallnor, Steve Raasch, and Steve Reinhardt, with contributions
 * from Ron Dreslinski, Dave Greene, Lisa Hsu, Kevin Lim, Ali Saidi,
 * and Andrew Schultz.
 *
 * Permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any
 * purpose, so long as the copyright notice above, this grant of
 * permission, and the disclaimer below appear in all copies made; and
 * so long as the name of The University of Michigan is not used in
 * any advertising or publicity pertaining to the use or distribution
 * of this software without specific, written prior authorization.
 *
 * THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 * UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND
 * WITHOUT WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. THE REGENTS OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE
 * LIABLE FOR ANY DAMAGES, INCLUDING DIRECT, SPECIAL, INDIRECT,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES, WITH RESPECT TO ANY CLAIM
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OF THE SOFTWARE, EVEN
 * IF IT HAS BEEN OR IS HEREAFTER ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGES.
 */

/**
 * @file
 * Miss and writeback queue definitions.
 */

#include "cpu/exec_context.hh"
#include "cpu/smt.hh" //for maxThreadsPerCPU
#include "mem/cache/base_cache.hh"
#include "mem/cache/miss/miss_queue.hh"
#include "mem/cache/prefetch/base_prefetcher.hh"

using namespace std;

// simple constructor
/**
 * @todo Remove the +16 from the write buffer constructor once we handle
 * stalling on writebacks do to compression writes.
 */
MissQueue::MissQueue(int numMSHRs, int numTargets, int write_buffers,
		     bool write_allocate, bool prefetch_miss)
    : mq(numMSHRs, 4), wb(write_buffers,numMSHRs+1000), numMSHR(numMSHRs), 
      numTarget(numTargets), writeBuffers(write_buffers), 
      writeAllocate(write_allocate), order(0), prefetchMiss(prefetch_miss)
{
    noTargetMSHR = NULL;
}

void
MissQueue::regStats(const string &name)
{
    using namespace Stats;

    writebacks
	.init(maxThreadsPerCPU)
	.name(name + ".writebacks")
	.desc("number of writebacks")
	.flags(total)
	;

    // MSHR hit statistics
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshr_hits[access_idx]
	    .init(maxThreadsPerCPU)
	    .name(name + "." + cstr + "_mshr_hits")
	    .desc("number of " + cstr + " MSHR hits")
	    .flags(total | nozero | nonan)
	    ;
    }

    demandMshrHits
	.name(name + ".demand_mshr_hits")
	.desc("number of demand (read+write) MSHR hits")
	.flags(total)
	;
    demandMshrHits = mshr_hits[Read] + mshr_hits[Write];

    overallMshrHits
	.name(name + ".overall_mshr_hits")
	.desc("number of overall MSHR hits")
	.flags(total)
	;
    overallMshrHits = demandMshrHits + mshr_hits[Soft_Prefetch] +
	mshr_hits[Hard_Prefetch];

    // MSHR miss statistics
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshr_misses[access_idx]
	    .init(maxThreadsPerCPU)
	    .name(name + "." + cstr + "_mshr_misses")
	    .desc("number of " + cstr + " MSHR misses")
	    .flags(total | nozero | nonan)
	    ;
    }

    demandMshrMisses
	.name(name + ".demand_mshr_misses")
	.desc("number of demand (read+write) MSHR misses")
	.flags(total)
	;
    demandMshrMisses = mshr_misses[Read] + mshr_misses[Write];

    overallMshrMisses
	.name(name + ".overall_mshr_misses")
	.desc("number of overall MSHR misses")
	.flags(total)
	;
    overallMshrMisses = demandMshrMisses + mshr_misses[Soft_Prefetch] +
	mshr_misses[Hard_Prefetch];

    // MSHR miss latency statistics
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshr_miss_latency[access_idx]
	    .init(maxThreadsPerCPU)
	    .name(name + "." + cstr + "_mshr_miss_latency")
	    .desc("number of " + cstr + " MSHR miss cycles")
	    .flags(total | nozero | nonan)
	    ;
    }

    demandMshrMissLatency
	.name(name + ".demand_mshr_miss_latency")
	.desc("number of demand (read+write) MSHR miss cycles")
	.flags(total)
	;
    demandMshrMissLatency = mshr_miss_latency[Read] + mshr_miss_latency[Write];

    overallMshrMissLatency
	.name(name + ".overall_mshr_miss_latency")
	.desc("number of overall MSHR miss cycles")
	.flags(total)
	;
    overallMshrMissLatency = demandMshrMissLatency +
	mshr_miss_latency[Soft_Prefetch] + mshr_miss_latency[Hard_Prefetch];

    // MSHR uncacheable statistics
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshr_uncacheable[access_idx]
	    .init(maxThreadsPerCPU)
	    .name(name + "." + cstr + "_mshr_uncacheable")
	    .desc("number of " + cstr + " MSHR uncacheable")
	    .flags(total | nozero | nonan)
	    ;
    }

    overallMshrUncacheable
	.name(name + ".overall_mshr_uncacheable_misses")
	.desc("number of overall MSHR uncacheable misses")
	.flags(total)
	;
    overallMshrUncacheable = mshr_uncacheable[Read] + mshr_uncacheable[Write] 
	+ mshr_uncacheable[Soft_Prefetch] + mshr_uncacheable[Hard_Prefetch];

    // MSHR miss latency statistics
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshr_uncacheable_lat[access_idx]
	    .init(maxThreadsPerCPU)
	    .name(name + "." + cstr + "_mshr_uncacheable_latency")
	    .desc("number of " + cstr + " MSHR uncacheable cycles")
	    .flags(total | nozero | nonan)
	    ;
    }

    overallMshrUncacheableLatency
	.name(name + ".overall_mshr_uncacheable_latency")
	.desc("number of overall MSHR uncacheable cycles")
	.flags(total)
	;
    overallMshrUncacheableLatency = mshr_uncacheable_lat[Read] 
	+ mshr_uncacheable_lat[Write] + mshr_uncacheable_lat[Soft_Prefetch]
	+ mshr_uncacheable_lat[Hard_Prefetch];

#if 0
    // MSHR access formulas
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshrAccesses[access_idx]
	    .name(name + "." + cstr + "_mshr_accesses")
	    .desc("number of " + cstr + " mshr accesses(hits+misses)")
	    .flags(total | nozero | nonan)
	    ;
	mshrAccesses[access_idx] =
	    mshr_hits[access_idx] + mshr_misses[access_idx]
	    + mshr_uncacheable[access_idx];
    }

    demandMshrAccesses
	.name(name + ".demand_mshr_accesses")
	.desc("number of demand (read+write) mshr accesses")
	.flags(total | nozero | nonan)
	;
    demandMshrAccesses = demandMshrHits + demandMshrMisses;

    overallMshrAccesses
	.name(name + ".overall_mshr_accesses")
	.desc("number of overall (read+write) mshr accesses")
	.flags(total | nozero | nonan)
	;
    overallMshrAccesses = overallMshrHits + overallMshrMisses
	+ overallMshrUncacheable;
#endif

    // MSHR miss rate formulas
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	mshrMissRate[access_idx]
	    .name(name + "." + cstr + "_mshr_miss_rate")
	    .desc("mshr miss rate for " + cstr + " accesses")
	    .flags(total | nozero | nonan)
	    ;

	mshrMissRate[access_idx] =
	    mshr_misses[access_idx] / cache->accesses[access_idx];
    }

    demandMshrMissRate
	.name(name + ".demand_mshr_miss_rate")
	.desc("mshr miss rate for demand accesses")
	.flags(total)
	;
    demandMshrMissRate = demandMshrMisses / cache->demandAccesses;

    overallMshrMissRate
	.name(name + ".overall_mshr_miss_rate")
	.desc("mshr miss rate for overall accesses")
	.flags(total)
	;
    overallMshrMissRate = overallMshrMisses / cache->overallAccesses;

    // mshrMiss latency formulas
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	avgMshrMissLatency[access_idx]
	    .name(name + "." + cstr + "_avg_mshr_miss_latency")
	    .desc("average " + cstr + " mshr miss latency")
	    .flags(total | nozero | nonan)
	    ;

	avgMshrMissLatency[access_idx] =
	    mshr_miss_latency[access_idx] / mshr_misses[access_idx];
    }

    demandAvgMshrMissLatency
	.name(name + ".demand_avg_mshr_miss_latency")
	.desc("average overall mshr miss latency")
	.flags(total)
	;
    demandAvgMshrMissLatency = demandMshrMissLatency / demandMshrMisses;

    overallAvgMshrMissLatency
	.name(name + ".overall_avg_mshr_miss_latency")
	.desc("average overall mshr miss latency")
	.flags(total)
	;
    overallAvgMshrMissLatency = overallMshrMissLatency / overallMshrMisses;

    // mshrUncacheable latency formulas
    for (int access_idx = 0; access_idx < NUM_MEM_CMDS; ++access_idx) {
	MemCmd cmd = (MemCmdEnum)access_idx;
	const string &cstr = cmd.toString();

	avgMshrUncacheableLatency[access_idx]
	    .name(name + "." + cstr + "_avg_mshr_uncacheable_latency")
	    .desc("average " + cstr + " mshr uncacheable latency")
	    .flags(total | nozero | nonan)
	    ;

	avgMshrUncacheableLatency[access_idx] =
	    mshr_uncacheable_lat[access_idx] / mshr_uncacheable[access_idx];
    }

    overallAvgMshrUncacheableLatency
	.name(name + ".overall_avg_mshr_uncacheable_latency")
	.desc("average overall mshr uncacheable latency")
	.flags(total)
	;
    overallAvgMshrUncacheableLatency = overallMshrUncacheableLatency / overallMshrUncacheable;

    mshr_cap_events
	.init(maxThreadsPerCPU)
	.name(name + ".mshr_cap_events")
	.desc("number of times MSHR cap was activated")
	.flags(total)
	;

    //software prefetching stats
    soft_prefetch_mshr_full
	.init(maxThreadsPerCPU)
	.name(name + ".soft_prefetch_mshr_full")
	.desc("number of mshr full events for SW prefetching instrutions")
	.flags(total)
	;

    mshr_no_allocate_misses
	.name(name +".no_allocate_misses")
	.desc("Number of misses that were no-allocate")
	;

}

void
MissQueue::setCache(BaseCache *_cache)
{
    cache = _cache;
    blkSize = cache->getBlockSize();
}

void
MissQueue::setPrefetcher(BasePrefetcher *_prefetcher)
{
    prefetcher = _prefetcher;
}

MSHR*
MissQueue::allocateMiss(MemReqPtr &req, int size, Tick time)
{
    
    MSHR* mshr = mq.allocate(req, size);
    
    if(cache->isDirectoryAndL1DataCache()){
        assert(mshr->directoryOriginalCmd == InvalidCmd);
        mshr->directoryOriginalCmd = req->cmd;
    }
    
    mshr->order = order++;
    if (!req->isUncacheable() ){//&& !req->isNoAllocate()) {
	// Mark this as a cache line fill
	mshr->req->flags |= CACHE_LINE_FILL;
    }
    if (mq.isFull()) {
	cache->setBlocked(Blocked_NoMSHRs);
    }
    if (req->cmd != Hard_Prefetch) {
	//If we need to request the bus (not on HW prefetch), do so
	cache->setMasterRequest(Request_MSHR, time);
    }
    return mshr;
}


MSHR*
MissQueue::allocateWrite(MemReqPtr &req, int size, Tick time)
{
    
    MSHR* mshr = wb.allocate(req,req->size);
    mshr->order = order++;
    if (cache->doData()){
	if (req->isCompressed()) {
	    delete [] mshr->req->data;
	    mshr->req->actualSize = req->actualSize;
	    mshr->req->data = new uint8_t[req->actualSize];
	    memcpy(mshr->req->data, req->data, req->actualSize);
	} else {
	    memcpy(mshr->req->data, req->data, req->size);
	}   
    }
    if (wb.isFull()) {
	cache->setBlocked(Blocked_NoWBBuffers);
    }

//     if(cache->getMasterInterfaceType() == CROSSBAR){
//         cache->setMasterRequestAddr(req->paddr);
//     }
    cache->setMasterRequest(Request_WB, time);

    return mshr;
}


/**
 * @todo Remove SW prefetches on mshr hits.
 */
void
MissQueue::handleMiss(MemReqPtr &req, int blkSize, Tick time)
{
    
//    if (!cache->isTopLevel())
    if (prefetchMiss) prefetcher->handleMiss(req, time);

    int size = blkSize;
    Addr blkAddr = req->paddr & ~(Addr)(blkSize-1);
    
    MSHR* mshr = NULL;
    if (!req->isUncacheable()) {
        
//         if(cache->isDirectoryAndL1DataCache()){
//             
//             // split MSHR organisations, a register either stores reads or it stores writes
//             mshr = mq.findMatch(blkAddr, req->cmd);
// //             if(cache->isDirectoryAndL1DataCache() && blkAddr == 5368749248ull){
// //                 if(mshr == NULL) cout << "did not find MSHR\n";
// //                 else cout << "found MSHR\n";
// //             }
//         }
//         else{
        mshr = mq.findMatch(blkAddr, req->asid);
//         }
        
	if (mshr){
//             if(cache->isDirectoryAndL1DataCache() && blkAddr == 5368749248ull){
//                 cout << "allocating as target\n";
//             }
            //@todo remove hw_pf here
            mshr_hits[req->cmd.toIndex()][req->thread_num]++;
            if (mshr->threadNum != req->thread_num) {
                mshr->threadNum = -1;
            }
            mq.allocateTarget(mshr, req);
            if (mshr->req->isNoAllocate() && !req->isNoAllocate()) {
                //We are adding an allocate after a no-allocate
                mshr->req->flags &= ~NO_ALLOCATE;
            }
            if (mshr->getNumTargets() == numTarget) {
                
//                 if(cache->isDirectoryAndL1DataCache() && blkAddr == 5368749248ull){
//                     cout << "blocking, no free targets\n";
//                 }
                
                noTargetMSHR = mshr;
                cache->setBlocked(Blocked_NoTargets);
                mq.moveToFront(mshr);
            }
            return;
            
	}
	if (req->isNoAllocate()) {
	    //Count no-allocate requests differently
	    mshr_no_allocate_misses++;
	}
	else {
	    mshr_misses[req->cmd.toIndex()][req->thread_num]++;
	}
    } else {
	//Count uncacheable accesses
	mshr_uncacheable[req->cmd.toIndex()][req->thread_num]++;
	size = req->size;
    }
    if (req->cmd.isWrite() && (req->isUncacheable() || !writeAllocate ||
			       req->cmd.isNoResponse())) {
	/**
	 * @todo Add write merging here.
	 */
	mshr = allocateWrite(req, req->size, time);
	return;
    }

    mshr = allocateMiss(req, size, time);
}

MSHR* 
MissQueue::fetchBlock(Addr addr, int asid, int blk_size, Tick time,
		      MemReqPtr &target)
{
    
    Addr blkAddr = addr & ~(Addr)(blk_size - 1);
    assert(mq.findMatch(addr, asid) == NULL);
    MSHR *mshr = mq.allocateFetch(blkAddr, asid, blk_size, target);
    mshr->order = order++;
    mshr->req->flags |= CACHE_LINE_FILL;
    if (mq.isFull()) {
	cache->setBlocked(Blocked_NoMSHRs);
    }
    
//     if(cache->getMasterInterfaceType() == CROSSBAR){
//         cache->setMasterRequestAddr(blkAddr);
//     }
    cache->setMasterRequest(Request_MSHR, time);
    return mshr;
}

MemReqPtr
MissQueue::getMemReq()
{
    MemReqPtr req = mq.getReq();
    
    if (((wb.isFull() && wb.inServiceMSHRs == 0) || !req || 
	 req->time > curTick) && wb.havePending()) {
	req = wb.getReq();
	// Need to search for earlier miss.
	MSHR *mshr = mq.findPending(req);
	if (mshr && mshr->order < req->mshr->order) {
	    // Service misses in order until conflict is cleared.
	    return mq.getReq();
	}
    }
    if (req) {
	MSHR* mshr = wb.findPending(req);
	if (mshr /*&& mshr->order < req->mshr->order*/) {
	    // The only way this happens is if we are
	    // doing a write and we didn't have permissions
	    // then subsequently saw a writeback(owned got evicted)
	    // We need to make sure to perform the writeback first
	    // To preserve the dirty data, then we can issue the write
	    return wb.getReq();
	}
    }
    else if (!mq.isFull()){
	//If we have a miss queue slot, we can try a prefetch
	req = prefetcher->getMemReq();
	if (req) {
	    //Update statistic on number of prefetches issued (hwpf_mshr_misses)
	    mshr_misses[req->cmd.toIndex()][req->thread_num]++;
	    //It will request the bus for the future, but should clear that immedieatley
	    allocateMiss(req, req->size, curTick);
	    req = mq.getReq();
	    assert(req); //We should get back a req b/c we just put one in
	}
    }
    
    return req;
}

void
MissQueue::setBusCmd(MemReqPtr &req, MemCmd cmd)
{
    
    assert(req->mshr != 0);
    MSHR * mshr = req->mshr;
    mshr->originalCmd = req->cmd;
    if (req->isCacheFill() || req->isNoAllocate())
	req->cmd = cmd;
}

void
MissQueue::restoreOrigCmd(MemReqPtr &req)
{
    req->cmd = req->mshr->originalCmd;
}

void
MissQueue::markInService(MemReqPtr &req)
{
    assert(req->mshr != 0);
    bool unblock = false;
    BlockedCause cause = NUM_BLOCKED_CAUSES;
    
    /**
     * @todo Should include MSHRQueue pointer in MSHR to select the correct
     * one.
     */
    if ((!req->isCacheFill() && req->cmd.isWrite()) || req->cmd == Copy) {
	// Forwarding a write/ writeback, don't need to change
	// the command
	unblock = wb.isFull();
	wb.markInService(req->mshr);
	if (!wb.havePending()){
	    cache->clearMasterRequest(Request_WB);
	}
	if (unblock) {
	    // Do we really unblock?
	    unblock = !wb.isFull();
	    cause = Blocked_NoWBBuffers;
	}
    } else {
	unblock = mq.isFull();
	mq.markInService(req->mshr);
	if (!mq.havePending()){
	    cache->clearMasterRequest(Request_MSHR);
	}
	if (req->mshr->originalCmd == Hard_Prefetch) {
	    DPRINTF(HWPrefetch, "%s:Marking a HW_PF in service\n", 
		    cache->name());
	    //Also clear pending if need be
	    if (!prefetcher->havePending())
	    {
                cache->clearMasterRequest(Request_PF);
	    }
	}
	if (unblock) {
	    unblock = !mq.isFull();
	    cause = Blocked_NoMSHRs;
	}
    }
    if (unblock) {
	cache->clearBlocked(cause);
    }
}


void
MissQueue::handleResponse(MemReqPtr &req, Tick time)
{

    MSHR* mshr = req->mshr;
    if (req->mshr->originalCmd == Hard_Prefetch) {
	DPRINTF(HWPrefetch, "%s:Handling the response to a HW_PF\n", 
		cache->name());
    }
    
#ifndef NDEBUG
    int num_targets = mshr->getNumTargets();
#endif
	
    bool unblock = false;
    bool unblock_target = false;
    BlockedCause cause = NUM_BLOCKED_CAUSES;
	
    if (req->isCacheFill() && !req->isNoAllocate()) {
	mshr_miss_latency[mshr->originalCmd][req->thread_num] +=
	    curTick - req->time;
	// targets were handled in the cache tags
	if (mshr == noTargetMSHR) {
	    // we always clear at least one target
	    unblock_target = true;
	    cause = Blocked_NoTargets;
	    noTargetMSHR = NULL;
	}
	
	if (mshr->hasTargets()) {
	    // Didn't satisfy all the targets, need to resend
	    MemCmd cmd = mshr->getTarget()->cmd;
	    mq.markPending(mshr, cmd);
	    mshr->order = order++;
            
            if(cache->isDirectoryAndL1DataCache()){
                // reset the addressing from the previous request
                req->toProcessorID = -1;
                req->fromProcessorID = -1;
                req->toInterfaceID = -1;
                req->fromInterfaceID = -1;
            }
            
	    cache->setMasterRequest(Request_MSHR, time);
	}
	else {
	    unblock = mq.isFull();
	    mq.deallocate(mshr);
	    if (unblock) {
		unblock = !mq.isFull();
		cause = Blocked_NoMSHRs;
	    }
	}
    } else {
        
	if (req->isUncacheable()) {
	    mshr_uncacheable_lat[req->cmd][req->thread_num] +=
		curTick - req->time;
	}
	if (mshr->hasTargets() && req->isUncacheable()) {
	    // Should only have 1 target if we had any
	    assert(num_targets == 1);
	    MemReqPtr target = mshr->getTarget();
	    mshr->popTarget();
	    if (cache->doData() && req->cmd.isRead()) {
		memcpy(target->data, req->data, target->size);
	    }
	    cache->respond(target, time);
	    assert(!mshr->hasTargets());
	}
	else if (mshr->hasTargets()) {
	    //Must be a no_allocate with possibly more than one target
	    assert(mshr->req->isNoAllocate());
	    while (mshr->hasTargets()) {
		MemReqPtr target = mshr->getTarget();
		mshr->popTarget();
		if (cache->doData() && req->cmd.isRead()) {
		    memcpy(target->data, req->data, target->size);
		}
		cache->respond(target, time);
       	    }
	}
	
	if (req->cmd.isWrite()) {
	    // If the wrtie buffer is full, we might unblock now
	    unblock = wb.isFull(); 
	    wb.deallocate(mshr);
	    if (unblock) {
		// Did we really unblock?
		unblock = !wb.isFull();
		cause = Blocked_NoWBBuffers;
	    }
	} else {
	    unblock = mq.isFull();
	    mq.deallocate(mshr);
	    if (unblock) {
		unblock = !mq.isFull();
		cause = Blocked_NoMSHRs;
	    }
	}
    }
    if (unblock || unblock_target) {
	cache->clearBlocked(cause);
    }
}

void
MissQueue::squash(int thread_number)
{
    
    bool unblock = false;
    BlockedCause cause = NUM_BLOCKED_CAUSES;

    if (noTargetMSHR && noTargetMSHR->threadNum == thread_number) {
	noTargetMSHR = NULL;
	unblock = true;
	cause = Blocked_NoTargets;
    }
    if (mq.isFull()) {
	unblock = true;
	cause = Blocked_NoMSHRs;
    }
    mq.squash(thread_number);
    if (!mq.havePending()) {
	cache->clearMasterRequest(Request_MSHR);
    }
    if (unblock && !mq.isFull()) {
	cache->clearBlocked(cause);
    }

}

MSHR*
MissQueue::findMSHR(Addr addr, int asid) const
{
    
    return mq.findMatch(addr,asid);
}

bool
MissQueue::findWrites(Addr addr, int asid, vector<MSHR*> &writes) const
{
    return wb.findMatches(addr,asid,writes);
}

void
MissQueue::doWriteback(Addr addr, int asid, ExecContext *xc,
		       int size, uint8_t *data, bool compressed)
{
    if(cache->useDirectory) fatal("Cache tags doWriteback() version not implemented with directory");
    
    // Generate request
    MemReqPtr req = buildWritebackReq(addr, asid, xc, size, data,
				      compressed);

    writebacks[req->thread_num]++;

    allocateWrite(req, 0, curTick);
}


void
MissQueue::doWriteback(MemReqPtr &req)
{
    writebacks[req->thread_num]++;
    assert(req->xc || !cache->doData());
    allocateWrite(req, 0, curTick);
}


MSHR* 
MissQueue::allocateTargetList(Addr addr, int asid)
{
   MSHR* mshr = mq.allocateTargetList(addr, asid, blkSize);
   mshr->req->flags |= CACHE_LINE_FILL;
   if (mq.isFull()) {
       cache->setBlocked(Blocked_NoMSHRs);
   }
   return mshr;
}

bool
MissQueue::havePending()
{
    return mq.havePending() || wb.havePending() || prefetcher->havePending();
}
