
#include "sim/builder.hh"
#include "crossbar.hh"
        
using namespace std;

void
Crossbar::request(Tick time, int fromID){
    
    requests++;
    
    if(isFirstRequest){
        //NOTE: This can not be initalised before alle interfaces has registered
        //      Consequently, doing it when the first request arrives is safe
        for(int i=0;i<allInterfaces.size();i++){
            requestQueues.push_back(new list<InterconnectRequest*>);
        }
        
        isFirstRequest = false;
    }
    
    insertIntoList(requestQueues[fromID], time, fromID);
    
    if(!blocked) scheduleArbitrationEvent(time + arbitrationDelay);

}

void
Crossbar::insertIntoList(list<InterconnectRequest* >* inList,
                         Tick reqTime, 
                         int fromID){

    if(inList->empty() || inList->back()->time <= reqTime){
        // fast common case;
        inList->push_back(new InterconnectRequest(reqTime, fromID));
    }
    else{
        list<InterconnectRequest*>::iterator pos;
        for(pos = inList->begin();
            pos != inList->end();
            pos++){
                if((*pos)->time > reqTime) break;
        }
        inList->insert(pos, new InterconnectRequest(reqTime, fromID));
    }
    
#ifdef DEBUG_CROSSBAR
    /* Make sure that the queues are sorted */
    for(int i=0;i<requestQueues.size();i++){
        assert(isSorted(requestQueues[i]));
    }
#endif //DEBUG_CROSSBAR
}

void
Crossbar::scheduleArbitrationEvent(Tick candidateTime){
    
    int found = false;
    for(int i=0;i<arbitrationEvents.size();i++){
        if(arbitrationEvents[i]->when() == candidateTime) found = true;
    }
    
    if(!found){
        InterconnectArbitrationEvent* event = 
                new InterconnectArbitrationEvent(this);
        event->schedule(candidateTime);
        arbitrationEvents.push_back(event);
    }
}

void
Crossbar::arbitrate(Tick cycle){
    
    assert(!blocked);
    
    Tick legalRequestTime = cycle - arbitrationDelay;
    
    /* create storage for the arbitraion process */
    vector<bool> deliverBusy(allInterfaces.size(), false);
    vector<bool> senderBusy(allInterfaces.size(), false);
    bool toOtherL1Busy = false;
    
    if(cycle < nextBusFreeTime) toOtherL1Busy = true;
    
    
    vector<list<InterconnectRequest*> > 
            delayedRequests(allInterfaces.size());
    for(int i=0;i<delayedRequests.size();i++){
        delayedRequests[i] = list<InterconnectRequest*>();
    }
    
    while(moreRequestsAvailable()){
        
        /* start with the oldest request, regardless of which queue it is in
           Consequently, starvation is avoided */
        int smallestId = -1;
        Tick smallest = TICK_T_MAX;
        for(int i=0;i<requestQueues.size();i++){
            if(!requestQueues[i]->empty() 
                && requestQueues[i]->front()->time < smallest){
                smallest = requestQueues[i]->front()->time;
                smallestId = i;
            }
        }
        assert(smallestId >= 0);
        
        InterconnectRequest* tempReq = requestQueues[smallestId]->front();
        requestQueues[smallestId]->pop_front();
        
        if(tempReq->time <= legalRequestTime){
        
            int toID = getDestinationId(tempReq->fromID);
            
            if(toID == -1){
                /* this was a null request, remove it and move on */
                delete tempReq;
                continue;
            }
            
            if(allInterfaces[toID]->isMaster()
               && allInterfaces[tempReq->fromID]->isMaster()){
                
                if(toOtherL1Busy){
                    delayedRequests[tempReq->fromID].push_back(tempReq);
                    continue;
                }
                else{
                    toOtherL1Busy = true;
                    nextBusFreeTime = (cycle + arbitrationDelay);
                    
                    /* update statistics */
                    arbitratedRequests++;
                    totalArbQueueCycles += 
                            (cycle - tempReq->time) - arbitrationDelay;
                    totalArbitrationCycles += arbitrationDelay;
                    
                    allInterfaces[tempReq->fromID]->grantData();
                    delete tempReq;
                    
                    continue;
                }
            }
            
            if(!deliverBusy[toID] && !senderBusy[tempReq->fromID]){
                /* request can be delivered*/
                
                /* update statistics */
                arbitratedRequests++;
                totalArbQueueCycles += 
                        (cycle - tempReq->time) - arbitrationDelay;
                totalArbitrationCycles += arbitrationDelay;
                
                allInterfaces[tempReq->fromID]->grantData();
                delete tempReq;
                
                
                deliverBusy[toID] = true;
                senderBusy[tempReq->fromID] = true;
                continue;
            }
        }
        delayedRequests[tempReq->fromID].push_back(tempReq);
    }
    
    /* put not granted requests back in the request queues */
    assert(requestQueues.size() == delayedRequests.size());
    for(int i=0;i<requestQueues.size();i++){
        assert(requestQueues[i]->empty());
        (*requestQueues[i]) = delayedRequests[i];
    }
    
    /* check if we need to schedule new arbitration events */
    Tick nextArbTime = TICK_T_MAX;
    bool addArbEvent = false;
    for(int i=0;i<requestQueues.size();i++){
        if(!requestQueues[i]->empty()){
            InterconnectRequest* first = requestQueues[i]->front();
            Tick candidateTime = first->time + arbitrationDelay;
            if(candidateTime <= cycle){
                if((cycle + 1) < nextArbTime){
                    addArbEvent = true;
                    nextArbTime = cycle + 1;
                }
            }
            else{
                if(candidateTime < nextArbTime){
                    addArbEvent = true;
                    nextArbTime = candidateTime;
                }
            }
        }
    }
    
    if(addArbEvent){
        assert(nextArbTime != TICK_T_MAX);
        scheduleArbitrationEvent(nextArbTime);
    }
}

bool
Crossbar::moreRequestsAvailable(){
    bool moreReqs = false;
    for(int i=0;i<requestQueues.size();i++){
        if(!requestQueues[i]->empty()) moreReqs = true;
    }
    return moreReqs;
}

int
Crossbar::getDestinationId(int fromID){
    
    if(allInterfaces[fromID]->isMaster()){
        
        pair<Addr,int> tmp = allInterfaces[fromID]->getTargetAddr();
        Addr targetAddr = tmp.first;
        int toInterfaceId = tmp.second;
        
        // The request was a null request, remove
        if(targetAddr == 0) return -1;
        
        // we allready know the to interface id if it's an L1 to L1 transfer
        if(toInterfaceId != -1) return toInterfaceId;
        
        return getTarget(targetAddr);
    }
    
    int retID = allInterfaces[fromID]->getTargetId();
    assert(retID != -1);
    return retID;
}

void
Crossbar::send(MemReqPtr& req, Tick time, int fromID){
    
    assert(!blocked);
    assert((req->size / width) <= 1);
    
    int toID = -1;
    bool busIsUsed = false;
    if(allInterfaces[fromID]->isMaster() && req->toInterfaceID != -1){
        busIsUsed = true;
        toID = req->toInterfaceID;
    }
    else if(allInterfaces[fromID]->isMaster()){
        toID = getTarget(req->paddr);
    }
    else{
        toID = req->fromInterfaceID;
    }
    
    //update profile stats
    if(doProfiling){
        if(busIsUsed){
            // the coherence bus is not pipelined
            channelUseCycles[allInterfaces.size()] += transferDelay;
        }
        else{
            // regular pipelined crossbar channels used
            // one pipeline slot is allocated in both sender
            // and recievers channel
            channelUseCycles[fromID] += 1;
            channelUseCycles[toID] += 1;
        }
    }
    
    deliverQueue.push_back(new InterconnectDelivery(time, fromID, toID, req));
    
    /* check if we need to schedule a deliver event */
    Tick deliverTime = time + transferDelay;
    bool found = false;
    for(int i=0;i<deliverEvents.size();i++){
        if(deliverEvents[i]->when() == deliverTime) found = true;
    }
    
    if(!found){
        InterconnectDeliverQueueEvent* event = 
                new InterconnectDeliverQueueEvent(this);
        event->schedule(deliverTime);
        deliverEvents.push_back(event);
    }
}


void
Crossbar::deliver(MemReqPtr& req, Tick cycle, int toID, int fromID){
    assert(!blocked);
    
    assert(!req);
    assert(toID == -1);
    assert(fromID == -1);
    
#ifdef DEBUG_CROSSBAR
    assert(isSorted(&deliverQueue));
#endif //DEBUG_CROSSBAR
    
    Tick legalGrantTime = cycle - transferDelay;

    /* attempt to deliver as many requests as possible */
    /* since the queue is sorted, starvation is not possible */
    while(!deliverQueue.empty()){
        InterconnectDelivery* delivery = deliverQueue.front();
        
        /* check if this grant has experienced the proper delay */
        /* since the requests are sorted, we know that all other are later */
        if(delivery->grantTime > legalGrantTime) break;
        
        deliverQueue.pop_front();
        
        /* update statistics */
        sentRequests++;
        int curCpuId = delivery->req->xc->cpu->params->cpu_id;
        int queueCycles = (cycle - delivery->grantTime) - transferDelay;
        
        totalTransQueueCycles += queueCycles;
        totalTransferCycles += transferDelay;
        perCpuTotalTransQueueCycles[curCpuId] += queueCycles;
        perCpuTotalTransferCycles[curCpuId] += transferDelay;
        
        
        int retval = BA_NO_RESULT;
        if(allInterfaces[delivery->toID]->isMaster()){
            allInterfaces[delivery->toID]->deliver(delivery->req);
        }
        else{
            retval = allInterfaces[delivery->toID]->access(delivery->req);
        }
        
        delete delivery;
        
        /* if the cache returns blocked we cannot deliver any more data */
        if(retval == BA_BLOCKED) break;
    }
}

void
Crossbar::setBlocked(int fromInterface){
    
    if(blocked) warn("blocking on a second cause");
    blocked = true;
    waitingFor = fromInterface;
    blockedAt = curTick;
    numSetBlocked++;
    
    blockedInterfaces.push_back(fromInterface);
    
    for(int i=0;i<arbitrationEvents.size();++i){
        if(arbitrationEvents[i]->scheduled()){
            arbitrationEvents[i]->deschedule();
        }
        delete arbitrationEvents[i];
    }
    arbitrationEvents.clear();
    
    for(int i=0;i<deliverEvents.size();++i){
        if(deliverEvents[i]->scheduled()){
            deliverEvents[i]->deschedule();
        }
        
        delete deliverEvents[i];
    }
    deliverEvents.clear();

}

void
Crossbar::clearBlocked(int fromInterface){
    
    assert(blocked);
    assert(blockedAt > -1);
        
    int hitIndex = -1;
    int hitCount = 0;
    for(int i=0;i<blockedInterfaces.size();i++){
        if(blockedInterfaces[i] == fromInterface){
            hitIndex = i;
            hitCount++;
        }
    }
    assert(hitCount == 1 && hitIndex > -1);
    blockedInterfaces.erase(blockedInterfaces.begin()+hitIndex);
    
    if(blockedInterfaces.empty()){
        
        blocked = false;
        numClearBlocked++;
        
        assert(arbitrationEvents.empty());
        if(moreRequestsAvailable()){
            
            /* schedule new arbitration event */
            Tick firstReq = TICK_T_MAX;
            for(int i=0;i<requestQueues.size();i++){
                if(!requestQueues[i]->empty() 
                    && requestQueues[i]->front()->time < firstReq){
                    firstReq = requestQueues[i]->front()->time;
                }
            }
            assert(firstReq != TICK_T_MAX);
            
            InterconnectArbitrationEvent* event = 
                    new InterconnectArbitrationEvent(this);
            arbitrationEvents.push_back(event);
            
            if(firstReq <= curTick) event->schedule(curTick);
            else event->schedule(firstReq);
        }
        
        assert(deliverEvents.empty());
        if(!deliverQueue.empty()){
            
            /* schedule new deliver event */
            Tick firstGrant = (deliverQueue.front())->grantTime;
            
            InterconnectDeliverQueueEvent* event = 
                    new InterconnectDeliverQueueEvent(this);
            deliverEvents.push_back(event);
            
            if(firstGrant <= curTick) event->schedule(curTick);
            else event->schedule(firstGrant);
        }
        blockedAt = -1;
    }
    
    
}

vector<int>
Crossbar::getChannelSample(){
    
    if(!doProfiling) doProfiling = true;
    
    std::vector<int> retval(channelUseCycles);
    
    for(int i=0;i<channelUseCycles.size();i++){
        channelUseCycles[i] = 0;
    }
    
    return retval;
}

void
Crossbar::writeChannelDecriptor(std::ofstream &stream){
    
    for(int i=0;i<allInterfaces.size();i++){
        stream << "Channel " << i << ": " 
               << allInterfaces[i]->getCacheName() << "\n";
    }
    stream << "Channel " << allInterfaces.size() << ": Coherence bus\n";
}

    
#ifdef DEBUG_CROSSBAR

void
Crossbar::printRequests(){
    for(int i=0;i<requestQueues.size();i++){
        cout << i << ": ";
        for(list<InterconnectRequest*>::iterator j=requestQueues[i]->begin();
            j != requestQueues[i]->end();
            j++){
                InterconnectRequest* current = *j;
                cout << "(" 
                        << current->fromID 
                        << ", " 
                        << current->time 
                        << ") ";
        }
        cout << "\n";
    }
    cout << "\n";
}

#endif //DEBUG_CROSSBAR

#ifndef DOXYGEN_SHOULD_SKIP_THIS

BEGIN_DECLARE_SIM_OBJECT_PARAMS(Crossbar)
    Param<int> width;
    Param<int> clock;
    Param<int> transferDelay;
    Param<int> arbitrationDelay;
    Param<int> cpu_count;
    SimObjectParam<HierParams *> hier;
END_DECLARE_SIM_OBJECT_PARAMS(Crossbar)

BEGIN_INIT_SIM_OBJECT_PARAMS(Crossbar)
    INIT_PARAM(width, "the width of the crossbar transmission channels"),
    INIT_PARAM(clock, "crossbar clock"),
    INIT_PARAM(transferDelay, "crossbar transfer delay in CPU cycles"),
    INIT_PARAM(arbitrationDelay, "crossbar arbitration delay in CPU cycles"),
    INIT_PARAM(cpu_count, "the number of CPUs in the system"),
    INIT_PARAM_DFLT(hier,
                    "Hierarchy global variables",
                    &defaultHierParams)
END_INIT_SIM_OBJECT_PARAMS(Crossbar)

CREATE_SIM_OBJECT(Crossbar)
{
    return new Crossbar(getInstanceName(),
                                 width,
                                 clock,
                                 transferDelay,
                                 arbitrationDelay,
                                 cpu_count,
                                 hier);
}

REGISTER_SIM_OBJECT("Crossbar", Crossbar)

#endif //DOXYGEN_SHOULD_SKIP_THIS
