
#include "sim/builder.hh"
#include "ideal_interconnect.hh"
        
using namespace std;

void
IdealInterconnect::request(Tick time, int fromID){
    
    requests++;
    
    // keep linked list of requests sorted at all times
    // first request takes priority over later requests at same cycle
    list<InterconnectRequest*>::iterator findPos;
    for(findPos=requestQueue.begin();
        findPos!=requestQueue.end();
        findPos++){
            
            InterconnectRequest* tempReq = *findPos;
            if(time < tempReq->time) break;
    }
    
    requestQueue.insert(findPos, new InterconnectRequest(time, fromID));
    
#ifdef DEBUG_IDEAL_INTERCONNECT
    
    /* check that the queue is sorted */
    InterconnectRequest* prev = NULL;
    bool first = true;
    for(list<InterconnectRequest*>::iterator i=requestQueue.begin();
        i!=requestQueue.end();
        i++){
            if(first){
                first = false;
                prev = *i;
                continue;
            }
            
            assert(prev->time <= (*i)->time);
            prev = *i;
    }
    
//     printRequestQueue();
#endif //DEBUG_IDEAL_INTERCONNECT
    
    /* add arbitration event if we are not blocked */
    if(!blocked){
        scheduleArbitrationEvent(time + arbitrationDelay);
    }
}

void
IdealInterconnect::send(MemReqPtr& req, Tick time, int fromID){
    
    assert(!blocked);
    assert((req->size / width) <= 1);
    
    bool isFromMaster = allInterfaces[fromID]->isMaster();
    
    if(req->toInterfaceID != -1){
        /* cache-to-cache transfer */
        assert(req->toProcessorID != -1);
        assert(req->toInterfaceID == getInterconnectID(req->toProcessorID));
        grantQueue.push_back(new InterconnectDelivery(time, 
                                                      fromID, 
                                                      req->toInterfaceID,
                                                      req));
    }
    else if(isFromMaster){
        /* reciever is a slave interface */
        int recvCount = 0;
        int recvID = -1;
        
        for(int i=0;i<allInterfaces.size();++i){
            if(allInterfaces[i]->isMaster()) continue;
            
            if(allInterfaces[i]->inRange(req->paddr)){
                recvCount++;
                recvID = i;
            }
        }
        
        /* check for errors */
        if(recvCount > 1){
            fatal("More than one supplier for address in IdealInterconnect");
        }
        if(recvCount != 1){
            fatal("No supplier for address in IdealInterconnect");
        }
        assert(recvID >= 0);
        
        grantQueue.push_back(new InterconnectDelivery(time, 
                                                      fromID, 
                                                      recvID,
                                                      req));
    }
    else{
        /* reciever is a master interface*/
        grantQueue.push_back(new InterconnectDelivery(time,
                                                      fromID,
                                                      req->fromInterfaceID,
                                                      req));
    }
    
#ifdef DEBUG_IDEAL_INTERCONNECT
    /* check that the queue is sorted */
    InterconnectDelivery* prev = NULL;
    bool first = true;
    for(list<InterconnectDelivery*>::iterator i=grantQueue.begin();
        i!=grantQueue.end();
        i++){
            if(first){
                first = false;
                prev = *i;
                continue;
            }
            
            assert(prev->grantTime <= (*i)->grantTime);
            prev = *i;
        }
#endif //DEBUG_IDEAL_INTERCONNECT
    
    bool found = false;
    for(int i=0;i<deliverEvents.size();i++){
        if(deliverEvents[i]->when() == (time + transferDelay)) found = true;
    }
    
    if(!found){
        InterconnectDeliverQueueEvent* event = 
                new InterconnectDeliverQueueEvent(this);
        event->schedule(time + transferDelay);
        deliverEvents.push_back(event);
    }

}

void
IdealInterconnect::arbitrate(Tick cycle){
    
    assert(!blocked);
    
    list<InterconnectRequest*> tempGrantQueue;
    
    InterconnectRequest* lastReq = requestQueue.back();
    if((lastReq->time + arbitrationDelay) <= cycle){
        //all requests can be issued
        tempGrantQueue.splice(tempGrantQueue.end(), requestQueue);
    }
    else{
        
        list<InterconnectRequest*>::iterator firstOldReq;
        for(firstOldReq = requestQueue.begin();
            firstOldReq != requestQueue.end();
            firstOldReq++){
                if(((*firstOldReq)->time + arbitrationDelay) > cycle) break;
        }
        
        tempGrantQueue.splice(tempGrantQueue.end(),
                              requestQueue,
                              requestQueue.begin(),
                              firstOldReq);
    }
    
    for(list<InterconnectRequest*>::iterator i = tempGrantQueue.begin();
        i != tempGrantQueue.end();
        i++){
            
            /* update statistics */
            arbitratedRequests++;
            totalArbQueueCycles += (cycle - (*i)->time) - arbitrationDelay;
            totalArbitrationCycles += arbitrationDelay;
            
            allInterfaces[(*i)->fromID]->grantData();
            delete *i;
    }
    
    if(!requestQueue.empty()){
        InterconnectRequest* firstReq = requestQueue.front();
        scheduleArbitrationEvent(firstReq->time + arbitrationDelay);
    }
}

void
IdealInterconnect::scheduleArbitrationEvent(Tick time){
    
    bool arbEventExists = false;
    for(int i=0;i<arbitrationEvents.size();++i){
        if(time == arbitrationEvents[i]->when()) arbEventExists = true;
    }

    if(!arbEventExists){
        InterconnectArbitrationEvent* event = 
                new InterconnectArbitrationEvent(this);
        event->schedule(time);
        arbitrationEvents.push_back(event);
    }
    
}

void
IdealInterconnect::deliver(MemReqPtr& req, Tick cycle, int toID, int fromID){
    
    assert(!blocked);
    
    assert(!req);
    assert(toID == -1);
    assert(fromID == -1);
    
    list<InterconnectDelivery* > grantsThisCycle;
    InterconnectDelivery* lastReq = grantQueue.back();
    if((lastReq->grantTime + transferDelay) <= cycle){
        //all requests can be issued
        grantsThisCycle.splice(grantsThisCycle.end(), grantQueue);
    }
    else{
        
        list<InterconnectDelivery*>::iterator firstOldReq;
        for(firstOldReq = grantQueue.begin();
            firstOldReq != grantQueue.end();
            firstOldReq++){
                if(((*firstOldReq)->grantTime + transferDelay) > cycle) break;
            }
        
            grantsThisCycle.splice(grantsThisCycle.end(),
                                   grantQueue,
                                   grantQueue.begin(),
                                   firstOldReq);
    }
    
    
    while(!grantsThisCycle.empty()){
        InterconnectDelivery* tempGrant = grantsThisCycle.front();
        grantsThisCycle.pop_front();
        
        /* update statistics */
        sentRequests++;
        int queueCycles = (cycle - tempGrant->grantTime) - transferDelay;
        int curCpuId = tempGrant->req->xc->cpu->params->cpu_id;
        
        totalTransQueueCycles += queueCycles;
        totalTransferCycles += transferDelay;
        perCpuTotalTransQueueCycles[curCpuId] += queueCycles;
        perCpuTotalTransferCycles[curCpuId] += transferDelay;
        
        int retval = BA_NO_RESULT;
        if(allInterfaces[tempGrant->toID]->isMaster()){
            allInterfaces[tempGrant->toID]->deliver(tempGrant->req);
        }
        else{
            retval = allInterfaces[tempGrant->toID]->access(tempGrant->req);
        }
        
        /* this delivery got through, free memory */
        delete tempGrant;
        
        /* if the cache returns blocked we cannot deliver any more data */
        if(retval == BA_BLOCKED) break;
    }

    if(!grantsThisCycle.empty()){
        grantQueue.splice(grantQueue.begin(),
                          grantsThisCycle,
                          grantsThisCycle.begin(),
                          grantsThisCycle.end());
    }
}

void
IdealInterconnect::setBlocked(int fromInterface){
    
    if(blocked) warn("blocking on a second cause");
    blocked = true;
    waitingFor = fromInterface;
    blockedAt = curTick;
    numSetBlocked++;
    
    blockedInterfaces.push_back(fromInterface);
    
    for(int i=0;i<arbitrationEvents.size();++i){
        if(arbitrationEvents[i]->scheduled()){
            arbitrationEvents[i]->deschedule();
        }
        delete arbitrationEvents[i];
    }
    arbitrationEvents.clear();
    
    for(int i=0;i<deliverEvents.size();++i){
        if(deliverEvents[i]->scheduled()){
            deliverEvents[i]->deschedule();
        }
        
        delete deliverEvents[i];
    }
    deliverEvents.clear();
    
}

void
IdealInterconnect::clearBlocked(int fromInterface){
    
    assert(blocked);
    assert(blockedAt > -1);
    
    int hitIndex = -1;
    int hitCount = 0;
    for(int i=0;i<blockedInterfaces.size();i++){
        if(blockedInterfaces[i] == fromInterface){
            hitIndex = i;
            hitCount++;
        }
    }
    assert(hitCount == 1 && hitIndex > -1);
    blockedInterfaces.erase(blockedInterfaces.begin()+hitIndex);
    
    if(blockedInterfaces.empty()){
        
        blocked = false;
        numClearBlocked++;
        
        assert(arbitrationEvents.empty());
        if(!requestQueue.empty()){
            
            /* schedule new arbitration event */
            Tick firstReq = (requestQueue.front())->time;
            
            InterconnectArbitrationEvent* event = 
                    new InterconnectArbitrationEvent(this);
            arbitrationEvents.push_back(event);
            
            if((firstReq + arbitrationDelay) <= curTick){
                event->schedule(curTick);
            }
            else{
                event->schedule(firstReq + arbitrationDelay);
            }
        }
        
        assert(deliverEvents.empty());
        if(!grantQueue.empty()){
            
            /* schedule new deliver event */
            Tick firstGrant = (grantQueue.front())->grantTime;
            
            InterconnectDeliverQueueEvent* event = 
                    new InterconnectDeliverQueueEvent(this);
            deliverEvents.push_back(event);
            
            if((firstGrant + transferDelay) <= curTick){
                event->schedule(curTick);
            }
            else{
                event->schedule(firstGrant + transferDelay);
            }
        }
        
        blockedAt = -1;
    }
}

#ifdef DEBUG_IDEAL_INTERCONNECT

void
IdealInterconnect::printRequestQueue(){
    cout << "ReqQueue: ";
    for(list<InterconnectRequest*>::iterator i = requestQueue.begin();
        i != requestQueue.end();
        i++){
            cout << "(" << (*i)->time << ", " << (*i)->fromID << ") ";
    }
    cout << "\n";
}

void
IdealInterconnect::printGrantQueue(){
    cout << "GrantQueue: ";
    for(list<InterconnectDelivery*>::iterator i = grantQueue.begin();
        i != grantQueue.end();
        i++){
            cout << "(" 
                    << (*i)->grantTime 
                    << ", " 
                    << (*i)->fromID 
                    << ", " 
                    << (*i)->toID 
                    << ") ";
    }
    cout << "\n";
}

#endif //DEBUG_IDEAL_INTERCONNECT

#ifndef DOXYGEN_SHOULD_SKIP_THIS

BEGIN_DECLARE_SIM_OBJECT_PARAMS(IdealInterconnect)
    Param<int> width;
    Param<int> clock;
    Param<int> transferDelay;
    Param<int> arbitrationDelay;
    Param<int> cpu_count;
    SimObjectParam<HierParams *> hier;
END_DECLARE_SIM_OBJECT_PARAMS(IdealInterconnect)

BEGIN_INIT_SIM_OBJECT_PARAMS(IdealInterconnect)
    INIT_PARAM(width, "ideal interconnect width, set this to the cache line "
                      "width"),
    INIT_PARAM(clock, "ideal interconnect clock"),
    INIT_PARAM(transferDelay, "ideal interconnect transfer delay in CPU "
                              "cycles"),
    INIT_PARAM(arbitrationDelay, "ideal interconnect arbitration delay in CPU "
                                 "cycles"),
    INIT_PARAM(cpu_count, "the number of CPUs in the system"),
    INIT_PARAM_DFLT(hier,
                    "Hierarchy global variables",
                    &defaultHierParams)
END_INIT_SIM_OBJECT_PARAMS(IdealInterconnect)

CREATE_SIM_OBJECT(IdealInterconnect)
{
    return new IdealInterconnect(getInstanceName(),
                                 width,
                                 clock,
                                 transferDelay,
                                 arbitrationDelay,
                                 cpu_count,
                                 hier);
}

REGISTER_SIM_OBJECT("IdealInterconnect", IdealInterconnect)

#endif //DOXYGEN_SHOULD_SKIP_THIS
